﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Runtime.Serialization;
using UnityEngine;

namespace UPP.Engine
{

    public abstract class PropertyBase : ISerializable
    {
        private string _name = "Unknown property";
        private string _desc = "No description available.";
        private string _category = "Basic Attributes";
        private string _machineName = "unknown";

        private bool _exposeToTemplates = true;
        private bool _exposeToInstances = true;
        private int _valueCount = 1;
        private string _propagation = "";

        public string Name { get { return _name; } set { _name = value; } }
        public string Description { get { return _desc; } set { _desc = value; } }
        public string MachineName { get { return _machineName; } set { _machineName = value; } }

        public string Category { get { return _category; } set { _category = value; } }
        public int ValueCount { get { return _valueCount; } set { _valueCount = value; } }
        public bool ExposeToTemplates { get { return _exposeToTemplates; } set { _exposeToTemplates = value; } }
        public bool ExposeToInstances { get { return _exposeToInstances; } set { _exposeToInstances = value; } }
        public string ValuePropagation { get { return _propagation; } set { _propagation = value; } }

        public string Data { get { return Serialize(); } set { Deserialize(value); } }

        protected abstract string Serialize();
        protected abstract void Deserialize(string str);

        public abstract string GetDisplayValue(bool shorten);
        public abstract void GUIForSetting(OGUI.Panel parent);
        public abstract void GUIForValue(OGUI.Panel parent);
        public abstract void GUIForValueSetPage(int index);
        public abstract void SaveGUIForSetting();

        public abstract string GenerateSourceAttributes();
        public abstract string GenerateSourceDeclaration(bool isClass);

        public static T FromSerialized<T> (string str) where T: new()
        {
            return Serializer.FromString<T>(str);
        }

        public static PropertyBase FromSerialized(string type, string str)
        {
            return (PropertyBase)Serializer.FromString(str, type);
        }

        public string ToSerialized<T>() where T : class
        {
            return Serializer.ToString<T>(this as T);
        }

        public string ToSerialized(string type)
        {
            return Serializer.ToString(this, type);
        }

        private static Type BasePropertyType(Type t)
        {
            if (t == null)
                return null;

            if (t.Name == "Property`1")
                return t.GetGenericArguments()[0];
            else
            {
                var p = t.BaseType;
                if (p == null)
                    return null;
                else
                    return BasePropertyType(p);
            }
        }

        public static PropertyBase Blend(PropertyRecord metaBase, PropertyRecord dataBase, bool useDefault = false)
        {
            if (metaBase.PropertyType != dataBase.PropertyType)
                return null;

            PropertyBase m = (PropertyBase)Serializer.FromString(metaBase.PropertySerialized, metaBase.PropertyType);
            PropertyBase d = (PropertyBase)Serializer.FromString(dataBase.PropertySerialized, dataBase.PropertyType);

            Type type = Type.GetType(dataBase.PropertyType);
            Type baseType = BasePropertyType(type);
            if (baseType == null)
                return null;
          //  Type arrType = baseType.MakeArrayType();

            System.Reflection.MethodInfo m_get = null;
            System.Reflection.MethodInfo m_set = null;
            Array g;
            if (useDefault)
            {
                m_get = type.GetMethod("GetDefault");
                object def = m_get.Invoke(Convert.ChangeType(d, type), null);

                int cnt = m.ValueCount;
                if (cnt == 0)
                    cnt = Convert.ToInt32(Settings.GetOrSet("UPP_inf_values_starting_count", "1"));

                g = (Array)Array.CreateInstance(typeof(string), cnt);
                for (int i = 0; i < g.Length; i++)
                    g.SetValue(def.ToString(), i);
            }
            else
            {
                m_get = type.GetMethod("Pack");
                g = (Array)m_get.Invoke(Convert.ChangeType(d, type), null);
            }
            m_set = type.GetMethod("Unpack");
            object[] args = new object[] { g };
            m_set.Invoke(m, args);
            return m;
        }

        public PropertyBase()
        {

        }

        public PropertyBase(SerializationInfo info, StreamingContext ctxt)
        {
            Deserialize(info.GetString("data"));
            _name = info.GetString("name");
            _machineName = info.GetString("machine");
            _desc = info.GetString("desc");
            _category = info.GetString("category");
            _valueCount = info.GetInt32("count");
            _exposeToTemplates = info.GetBoolean("exTemp");
            _exposeToInstances = info.GetBoolean("exInst");
            _propagation = info.GetString("propagation");
        }

        public void GetObjectData(SerializationInfo info, StreamingContext ctxt)
        {
            info.AddValue("data", Serialize());
            info.AddValue("name", _name);
            info.AddValue("machine", _machineName);
            info.AddValue("desc", _desc);
            info.AddValue("category", _category);
            info.AddValue("count", _valueCount);
            info.AddValue("exTemp", _exposeToTemplates);
            info.AddValue("exInst", _exposeToInstances);
            info.AddValue("propagation", _propagation);
        }
    }

}