﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using UnityEditor;
using UnityEngine;
using UPP.OGUI;

namespace UPP.Engine
{
    public class TemplateEditor : Window
    {
        [MenuItem("Unity++/Template Editor")]
        public static void ShowWindow()
        {
            EditorWindow.GetWindow(typeof(TemplateEditor),
                Boolean.Parse(Settings.GetOrSet("UPP_windows_as_utility", true.ToString())),
                "Template Editor"
                );
        }

        const string CATEGORY_PREFIX = "__CATEGORY__@";

        TreeView hierarchyTreeView = null;
        Label infoLabel = null;
        Panel attributeCanvas = null;
        Table attributeTable = null;
        ComboBox templateTypesCombo = null;
        Panel tools = null;
        Button bAdd, bEdit, bRemove, aEdit;

        ClassRecord SelectedTemplateType = null;

        TemplateEditorEdit tempEditorEdit = null;

        void FillTemplateCombo()
        {
            templateTypesCombo.Clear();
            var all = ClassManager.DB.database;
            List<ClassRecord> list = new List<ClassRecord>();
            foreach (var c in all)
            {
                var cl = ClassManager.GetClassByMachineName(c.machineName);
                if (!cl.isAbstract)
                    list.Add(cl);
            }
            list.Sort((x, y) => x.className.CompareTo(y.className));
            foreach (var l in list)
            {
                templateTypesCombo.Add(l.machineName, l.className);
            }
        }

        void ReconstructTree()
        {
            hierarchyTreeView.Clear();
            if (SelectedTemplateType == null)
                return;

            var list = TemplateManager.FindTemplatesOfType(SelectedTemplateType.machineName);
            var cat = new List<KeyValuePair<string, List<TemplateRecord>>>();
            var cats = new List<string>();
            foreach (var l in list)
            {
                var rec = l.Value;
                if (!cats.Contains(rec.Category))
                {
                    cats.Add(rec.Category);
                    cat.Add(new KeyValuePair<string, List<TemplateRecord>>(rec.Category, new List<TemplateRecord>()));
                }
                cat.Find(x => x.Key == rec.Category).Value.Add(rec);
            }
            cat.Sort((x, y) => x.Key.CompareTo(y.Key));
            foreach (var c in cat)
            {
                var recs = c.Value;
                recs.Sort((x, y) => x.TemplateName.CompareTo(y.TemplateName));
                TreeView.TreeViewItem parent = null;
                if (c.Key != "")
                {
                    parent = new TreeView.TreeViewItem(CATEGORY_PREFIX + c.Key, c.Key);
                    hierarchyTreeView.AddItem(parent);
                }

                foreach (var rec in recs)
                {
                    TreeView.TreeViewItem item = new TreeView.TreeViewItem(rec.TemplateMachineName, rec.TemplateName);
                    hierarchyTreeView.AddItem(item, parent);
                }
            }
            hierarchyTreeView.Refresh();
        }

        void SetButtons(string name)
        {
            bAdd.Text = "New " + name;
            bEdit.Text = "Edit " + name;
            bRemove.Text = "Remove " + name;
            tools.Size.Width = bAdd.Size.Width + bRemove.Size.Width + bEdit.Size.Width + aEdit.Size.Width + 25;
            bAdd.Position = new Position(3, 3);
            bEdit.Position = new Position(bAdd.Position.X + bAdd.Size.Width + 3, 3);
            bRemove.Position = new Position(bEdit.Position.X + bEdit.Size.Width + 3, 3);
            aEdit.Position = new Position(bRemove.Position.X + bRemove.Size.Width + 13, 3);
            bAdd.Visible = true;
            bEdit.Visible = true;
            bRemove.Visible = true;
            aEdit.Visible = true;
        }

        protected override void PrepareGUI()
        {
            Panel menu = new Panel();
            menu.Dock = DockStyle.Top;
            menu.Size.Height = 25;
            menu.BorderColor = ColorSchemes.BorderColorDarker;

            Panel info = new Panel();
            info.Dock = DockStyle.Bottom;
            info.Size.Height = 25;
            info.BorderColor = ColorSchemes.BorderColorDarker;

            infoLabel = new Label();
            infoLabel.Parent = info;
            infoLabel.Size.Width = info.Size.Width - 16;
            infoLabel.Anchor = new Anchor(true, true, true, true);
            infoLabel.Position = new Position(8, 5);

            SplitContainer spl = new SplitContainer(SplitContainer.ModeType.Horizontal, 25, SplitContainer.SizeType.Percentage, SplitContainer.PanelWho.First);
            spl.Dock = DockStyle.Fill;

            attributeCanvas = new Panel();
            attributeCanvas.Parent = spl.GetPanel(SplitContainer.PanelWho.Second);
            attributeCanvas.Dock = DockStyle.Fill;
            attributeCanvas.Size = spl.GetPanel(SplitContainer.PanelWho.Second).Size;
            attributeCanvas.BackgroundColor = ColorSchemes.BackgroundColorDarker;
            attributeCanvas.SizeMode = Panel.SizeModeType.Clip;

            attributeTable = new Table(new string[] { "Attribute", "Type", "Value" });
            attributeTable.Parent = attributeCanvas;
            attributeTable.Columns[0].Width = 200;
            attributeTable.Columns[1].Width = 100;
            attributeTable.Size.Width = attributeCanvas.Size.Width;
            attributeTable.Anchor = new Anchor(true, true, true, true);
            attributeTable.BackgroundColor = ColorSchemes.BackgroundColor;
            attributeTable.OnDoubleClick += AttributeTable_OnDoubleClick;

            Panel hierarchy = new Panel();
            hierarchy.Parent = spl.GetPanel(SplitContainer.PanelWho.First);
            hierarchy.Dock = DockStyle.Fill;

            hierarchyTreeView = new TreeView();
            hierarchyTreeView.Parent = hierarchy;
            hierarchyTreeView.Dock = DockStyle.Fill;
            ReconstructTree();
            hierarchyTreeView.Refresh();
            hierarchyTreeView.OnSelectionChange += HierarchyTreeView_OnSelectionChange;
            hierarchyTreeView.OnDoubleClick += HierarchyTreeView_OnDoubleClick;
            hierarchyTreeView.ExpandAll();

            Label l = new Label("Template type:", true);
            l.Parent = menu;
            l.Position = new Position(8, 4);

            templateTypesCombo = new ComboBox("<Select Type>");
            templateTypesCombo.Parent = menu;
            templateTypesCombo.Position = new Position(120, 2);
            FillTemplateCombo();
            templateTypesCombo.OnChange += TemplateTypesCombo_OnChange;

            bAdd = new Button("Add");
            bAdd.Anchor = new Anchor(true, true, false, false);
            bAdd.Visible = false;
            bAdd.OnClick += BAdd_OnClick;

            bEdit = new Button("Edit");
            bEdit.Anchor = new Anchor(true, true, false, false);
            bEdit.Visible = false;
            bEdit.OnClick += BEdit_OnClick;

            bRemove = new Button("Remove");
            bRemove.Anchor = new Anchor(true, true, false, false);
            bRemove.Visible = false;
            bRemove.OnClick += BRemove_OnClick;

            aEdit = new Button("Edit Attribute");
            aEdit.Anchor = new Anchor(true, true, false, false);
            aEdit.Visible = false;
            aEdit.OnClick += AEdit_OnClick;

            tools = new Panel(menu);
            tools.Dock = DockStyle.Right;

            bAdd.Parent = tools;
            bEdit.Parent = tools;
            bRemove.Parent = tools;
            aEdit.Parent = tools;

            OnModalReturn += TemplateEditor_OnModalReturn;
        }

        private void AttributeTable_OnDoubleClick(object sender, int row)
        {
            EditAttribute();
        }

        private void AEdit_OnClick(object sender)
        {
            EditAttribute();
        }

        private void HierarchyTreeView_OnSelectionChange(object sender, string machineName, string displayName, int internalId)
        {
            SelectedTemplate();
        }

        private void BRemove_OnClick(object sender)
        {
            RemoveTemplate();
        }

        private void HierarchyTreeView_OnDoubleClick(object sender, string machineName, string displayName, int internalId)
        {
            EditTemplate();
        }

        private void BEdit_OnClick(object sender)
        {
            EditTemplate();
        }

        private void TemplateEditor_OnModalReturn(DialogResult result, Type childType, object child)
        {
            if (childType == typeof(TemplateEditorEdit))
            {
                if (result != DialogResult.Cancel)
                {
                    ReconstructTree();
                    hierarchyTreeView.SelectByMychineName(TemplateManager.LastManipulatedTemplateName);
                }
            }

            if (childType == typeof(PropertyWindowInstanceValues))
            {
                AssetDatabase.Refresh();

                lastEditedProperty.PropertySerialized = lastEditedPropertyBase.ToSerialized(lastEditedProperty.PropertyType);

                EditorUtility.SetDirty(lastEditedProperty);
                EditorUtility.SetDirty(lastEditedTemplate);
                AssetDatabase.SaveAssets();
                LoadAttributes(lastEditedTemplate);
            }
        }

        private void BAdd_OnClick(object sender)
        {
            AddTemplate();
        }

        private string CurrentCategory()
        {
            string cat = "";


            if (hierarchyTreeView.SelectedItem != null)
            {
                string m = hierarchyTreeView.SelectedItem.MachineName;
                if (m.IndexOf(CATEGORY_PREFIX) == 0)
                    cat = m.Substring(CATEGORY_PREFIX.Length);
                else
                {
                    var r = TemplateManager.GetTemplateByMachineName(m);
                    if (r != null && r.Category != "")
                        cat = r.Category;
                }
            }

            return cat;
        }

        private void AddTemplate()
        {
            if (SelectedTemplateType != null)
            {
                tempEditorEdit = EditorWindow.GetWindow(typeof(TemplateEditorEdit), true, "Create " + SelectedTemplateType.className) as TemplateEditorEdit;
                ModalChild = tempEditorEdit;
                tempEditorEdit.TemplateRec = null;
                tempEditorEdit.ParentClass = SelectedTemplateType.machineName;
                tempEditorEdit.PreCategory = CurrentCategory();
                tempEditorEdit.Mode = TemplateEditorEdit.EditorEditMode.Create;
            }
        }

        private void EditTemplate()
        {
            if (SelectedTemplateType != null && hierarchyTreeView.SelectedItem != null)
            {
                var machine = hierarchyTreeView.SelectedItem.MachineName;
                var cls = TemplateManager.GetTemplateByMachineName(machine);
                if (cls == null)
                    return;

                tempEditorEdit = EditorWindow.GetWindow(typeof(TemplateEditorEdit), true, "Edit " + SelectedTemplateType.className) as TemplateEditorEdit;
                ModalChild = tempEditorEdit;
                tempEditorEdit.TemplateRec = cls;
                tempEditorEdit.ParentClass = SelectedTemplateType.machineName;
                tempEditorEdit.Mode = TemplateEditorEdit.EditorEditMode.Edit;
            }
        }

        private void RemoveTemplate()
        {
            if (SelectedTemplateType != null && hierarchyTreeView.SelectedItem != null)
            {
                var machine = hierarchyTreeView.SelectedItem.MachineName;
                
                if (TemplateManager.Remove(machine, true))
                {
                    ReconstructTree();
                    hierarchyTreeView.ExpandAll();
                }
            }
        }

        private void TemplateTypesCombo_OnChange(object sender)
        {
            var machine = templateTypesCombo.SelectedKey;
            var cls = ClassManager.GetClassByMachineName(machine);
            SwitchTempType(cls);
        }

        private void SwitchTempType(ClassRecord cls)
        {
            if (cls != null)
            {
                SelectedTemplateType = cls;
                SetButtons(cls.className);
            }
            else
            {
                SelectedTemplateType = null;
                bAdd.Visible = false;
                bEdit.Visible = false;
                bRemove.Visible = false;
                aEdit.Visible = false;
            }
            ReconstructTree();
            hierarchyTreeView.ExpandAll();
            LoadAttributes(null);
        }

        private void LoadAttributes(TemplateRecord templ)
        {
            attributeTable.Clear();
            if (templ == null)
                return;

            Dictionary<string, List<PropertyBase>> list = new Dictionary<string, List<PropertyBase>>();

            foreach (var t in templ.PropertyRecords)
            {
                PropertyBase pr = PropertyBase.FromSerialized(t.PropertyType, t.PropertySerialized);
                if (pr.ExposeToTemplates)
                {
                    if (!list.ContainsKey(pr.Category))
                    {
                        list.Add(pr.Category, new List<PropertyBase>());
                    }
                    list[pr.Category].Add(pr);
                }
            }

            var categories = list.Keys.ToList();
            categories.Sort();

            foreach (var category in categories)
            {
                var batch = list[category];
                batch.Sort((x, y) => x.Name.CompareTo(y.Name));
                attributeTable.AddTitle(category);

                foreach (var b in batch)
                {
                    List<string> row = new List<string>();
                    PropertyInfo prInfo = (PropertyInfo)Attribute.GetCustomAttribute(b.GetType(), typeof(PropertyInfo));
                    row.Add(b.Name);
                    row.Add(prInfo.Name);
                    row.Add(b.GetDisplayValue(true));
                    attributeTable.AddRow(row);
                    attributeTable.GetLastRow().Tag = b.MachineName;
                }
            }
        }

        private void SelectedTemplate()
        {
            if (SelectedTemplateType != null && hierarchyTreeView.SelectedItem != null)
            {
                var templ = TemplateManager.GetTemplateByMachineName(hierarchyTreeView.SelectedItem.MachineName);
                if (templ == null)
                {
                    LoadAttributes(null);
                    return;
                }

                TemplateManager.PrepareAttributes(templ);
                LoadAttributes(templ);
            }
        }

        PropertyRecord lastEditedProperty = null;
        TemplateRecord lastEditedTemplate = null;
        PropertyBase lastEditedPropertyBase = null;

        private void EditAttribute()
        {
            if (SelectedTemplateType != null && hierarchyTreeView.SelectedItem != null && attributeTable.SelectedRow >= 0)
            {
                string machineProperty = attributeTable.Rows[attributeTable.SelectedRow].Tag;
                string propName = attributeTable.Rows[attributeTable.SelectedRow].Cells[0];

                var machineClass = hierarchyTreeView.SelectedItem.MachineName;
                var cls = TemplateManager.GetTemplateByMachineName(machineClass);
                if (cls == null)
                    return;

                var props = cls.PropertyRecords.Where(x => x.PropertyMachineName == machineProperty);
                PropertyRecord prop = null;
                if (props.Count() == 1)
                    prop = props.ElementAt(0);
                if (prop == null)
                    return;

                PropertyBase propBase = PropertyBase.FromSerialized(prop.PropertyType, prop.PropertySerialized);

                lastEditedProperty = prop;
                lastEditedTemplate = cls;
                lastEditedPropertyBase = propBase;

                var window = EditorWindow.GetWindow(typeof(PropertyWindowInstanceValues), true, cls.TemplateName + " - " + propName) as PropertyWindowInstanceValues;
                ModalChild = window;
                window.DeserializedProp = propBase;
            }
        }
    }

}